"""
Script to evaluate predictions.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import Dict, List, Optional
import os
import pandas as pd
import numpy as np
from sklearn.calibration import calibration_curve
import matplotlib.pyplot as plt
from matplotlib.axes import Axes

from ..utils.preprocessing import parse_path, create_dir
from ..utils.classification import metrics_from_predictions

import argparse


def abline(slope: float, intercept: float, axes: Axes) -> None:
    """
    Plot a line from slope and intercept on current axis.
    """
    x_vals = np.array(axes.get_xlim())
    y_vals = intercept + slope * x_vals
    axes.plot(x_vals, y_vals, linestyle='dotted', color='k')

def draw_reliability_diagram(
    y_true: np.ndarray, 
    y_probs: np.ndarray, 
    save_path: str, 
    method_name: str
    ) -> None:
    """
    Draws reliability diagram into save_path.
    save_path must not contain extension.
    """
    prob_true, prob_pred = calibration_curve(y_true, y_probs, n_bins=20)

    fig, ax = plt.subplots(1,2, figsize=(16,6))
    ax[0].plot(prob_pred, prob_true, marker='s', color='forestgreen')
    abline(1, 0, ax[0])
    ax[0].set(
        title='Reliability diagram', 
        xlabel='Mean predicted probability', 
        ylabel='Fraction of positives'
    )
    ax[0].legend([method_name, 'Perfectly calibrated'])

    ax[1].hist(
        y_probs,
        range=(0, 1),
        bins=20,
        label=method_name,
        color='forestgreen',
    )
    ax[1].set(title=method_name, xlabel="Mean predicted probability", ylabel="Count")

    fig.savefig(save_path + '.png')
    plt.close(fig)

def compute_metrics(
    nodes_df: pd.DataFrame, 
    draw_on: Optional[str] = None, 
    method_name: Optional[str] = 'Method'
    ) -> Dict[str, float]:
    """
    Computes F1-score, Accuracy, ROC AUC, Expected Calibration Error and percentage error.
    Dataframe must contain columns class and prob1.
    Class columns must have 1 for negative and 2 for positive.
    """
    labels = nodes_df['class'].to_numpy()-1
    probs = nodes_df['prob1'].to_numpy()
    preds = (probs > 0.5) * 1

    acc, f1, auc, perc_error, ece = metrics_from_predictions(labels, preds, probs, 2)

    if draw_on is not None:
        draw_reliability_diagram(labels, probs, draw_on, method_name)

    return {'Accuracy': acc, 'F1-score': f1, 'ROC AUC': auc, 'Error percentage': perc_error, 'ECE': ece}

def save_metrics(metrics: Dict[str, float], save_name):
    """
    Saves metrics into csv file.
    """
    metrics_df = pd.DataFrame(metrics)
    metrics_df.to_csv(save_name + '.csv', index=False)
    
def join_dictionaries(dict1: Dict[str, List[float]], dict2: Dict[str, float]) -> None:
    """
    Joins dict2 into dict1, modifying dict1.
    For each key it adds the value in dict2 to the list in dict1,
    if dict1 doesn't have such key, it is created.
    """
    for key in dict2.keys():
        if key not in dict1:
            dict1[key] = []
        dict1[key].append(dict2[key])


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--node-dir', type=str, required=True,
                        help='Folder with .nodes.csv prediction files.')
    parser.add_argument('--save-file', type=str, required=True,
                        help='Name to file where to save the results. Must not contain extension.')
    parser.add_argument('--by-img', action='store_true',
                        help='Whether to evaluate images separately or totally.')
    parser.add_argument('--draw', action='store_true',
                        help='Whether to draw reliability diagrams.')
    parser.add_argument('--draw-dir', type=str, 
                        help='Folder to save reliability diagram on by-img mode.')
    parser.add_argument('--name', type=str, default='Method', 
                        help='The name of the model used. Default: Method.')
    return parser


def main_with_args(args):
    node_files = os.listdir(args.node_dir)
    node_paths = [parse_path(args.node_dir) + x for x in node_files]
    metrics = {}
    if args.by_img:
        for node_path, node_file in zip(node_paths, node_files):
            nodes_df = pd.read_csv(node_path)
            if args.draw:
                assert args.draw_dir is not None, 'In by-img mode you must provide a folder for saving drawings.'
                draw_dir = parse_path(args.draw_dir)
                create_dir(draw_dir)
                aux_metrics = compute_metrics(nodes_df, draw_on=draw_dir + node_file, method_name=args.name)
            else:
                aux_metrics = compute_metrics(nodes_df)
            aux_metrics['name'] = node_file[:-10]
            join_dictionaries(metrics, aux_metrics)
    else:
        total_nodes_df = pd.DataFrame()
        for node_path in node_paths:
            nodes_df = pd.read_csv(node_path)
            total_nodes_df = pd.concat([total_nodes_df,nodes_df])
        total_nodes_df.reset_index(inplace=True)
        total_nodes_df.drop('index', axis=1, inplace=True)
        if args.draw:
            aux_metrics = compute_metrics(total_nodes_df, draw_on=args.save_file, method_name=args.name)
        else:
            aux_metrics = compute_metrics(total_nodes_df)
        join_dictionaries(metrics, aux_metrics)
    save_metrics(metrics, args.save_file)


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



from .centroids2png import main_with_args as centroids2png
from .geojson2pngcsv import main_with_args as geojson2pngcsv
from .hovernet2centroids import main_with_args as hovernet2centroids
from .hovernet2geojson import main_with_args as hovernet2geojson
from .pngcsv2centroids import main_with_args as pngcsv2centroids
from .pngcsv2geojson import main_with_args as pngcsv2geojson
from .pngcsv2graph import main_with_args as pngcsv2graph
from .graph2centroids import main_with_args as graph2centroids
from .centroidspng2csv import main_with_args as centroidspng2csv
from .png2graph import main_with_args as png2graph



"""
Extract HoVer-Net probabilities from json and
concatenates the result as new column (prob1) to .nodes.csv.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from tqdm import tqdm
from typing import Dict, Any, List, Tuple, Optional
import pandas as pd
import numpy as np
import argparse
from argparse import Namespace
import os
from ..utils.preprocessing import create_dir, parse_path, get_names, read_json, save_graph
from ..utils.nearest import generate_tree, find_nearest
import logging
from logging import Logger


def parse_centroids_probs(nuc: Dict[str, Any], logger: Optional[Logger] = None, num_classes: Optional[int] = 2) -> List[Tuple[int,int,int]]:
    """
    Input: Hovernet json nuclei dictionary as given by modified run_infer.py.
    Output: List of (X,Y,prob1) tuples representing centroids.
    """
    centroids_ = []
    for inst in nuc:
        inst_info = nuc[inst]
        inst_centroid = inst_info['centroid']
        inst_prob1 = inst_info['prob1']
        if num_classes > 2:
            inst_probs = []
            for k in range(1, num_classes + 1):
                inst_probs.append(inst_info['prob' + str(k)])
        inst_type = inst_info['type']
        if inst_type == 0:
            if logger is None:
                logging.warning('Found cell with class 0, removing it.')
            else:
                logger.warning('Found cell with class 0, removing it.')
        else:
            if num_classes == 2:
                centroids_.append((inst_centroid[1], inst_centroid[0], inst_prob1)) 
            else:
                centroids_.append((inst_centroid[1], inst_centroid[0], *inst_probs)) 
    return centroids_


def add_probability(graph: pd.DataFrame, hov_json: Dict[str, Any], logger: Optional[Logger] = None, num_classes: Optional[int] = 2) -> pd.DataFrame:
    """
    Extracts type_prob from json and adds it as column prob1.
    Makes the join based on id.
    """
    centroids = parse_centroids_probs(hov_json, logger, num_classes)
    centroids = np.array(centroids)
    assert len(centroids) > 0, 'Hov json must contain at least one cell.'
    graph = graph.copy()
    if not 'prob1' in graph.columns:
        n_cols = len(graph.columns)
        graph.insert(n_cols, 'prob1', [-1] * len(graph))
    else:
        graph['prob1'] = -1
    if num_classes > 2:
        for k in range(2, num_classes + 1):
            if not 'prob' + str(k) in graph.columns:
                n_cols = len(graph.columns)
                graph.insert(n_cols, 'prob' + str(k), [-1] * len(graph))
            else:
                graph['prob' + str(k)] = -1
    gt_tree = generate_tree(centroids[:,:2])
    pred_centroids = graph[['X', 'Y']].to_numpy(dtype=int)
    pred_tree = generate_tree(pred_centroids)
    for point_id, point in enumerate(centroids):
        closest_id = find_nearest(point[:2], pred_tree)
        closest = graph.loc[closest_id, ['X', 'Y', 'prob1']]
        if point_id == find_nearest(closest[:2], gt_tree):
            graph.loc[closest_id, 'prob1'] = point[2] # 1-1 matchings
            if num_classes > 2:
                for k in range(2, num_classes + 1):
                    graph.loc[closest_id, 'prob' + str(k)] = point[k + 1] # 1-1 matchings
    graph.drop(graph[graph['prob1'] == -1].index, inplace=True)
    return graph


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--json-dir', type=str, required=True,
        help='Path to folder containing HoVer-Net json outputs.'
    )
    parser.add_argument(
        '--graph-dir', type=str, required=True,
        help='Path to directory to .nodes.csv containing graph information.'
    )
    parser.add_argument(
        '--output-dir', type=str, required=True,
        help='Path where to save new .nodes.csv. If same as --graph-dir, overwrites its content.'
    )
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    return parser


def main_with_args(args: Namespace, logger: Optional[Logger] = None) -> None:
    json_dir = parse_path(args.json_dir)
    graph_dir = parse_path(args.graph_dir)
    output_dir = parse_path(args.output_dir)
    create_dir(output_dir)
    names = get_names(graph_dir, '.nodes.csv')
    for name in tqdm(names):
        try:
            graph = pd.read_csv(os.path.join(graph_dir, name + '.nodes.csv'))
            hov_json = read_json(os.path.join(json_dir, name + '.json'))
            graph = add_probability(graph, hov_json, logger, args.num_classes)
            save_graph(graph, os.path.join(output_dir, name + '.nodes.csv'))
        except FileNotFoundError:
            continue
    return


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



from argparse import Namespace
from logging import Logger
import os
import pandas as pd
from typing import List


class WrongConfigurationError(Exception):
    pass


class HovernetNotFoundError(Exception):
    pass


def check_void(path: str) -> bool:
    """Tells if a directory if void or if file exists. True if void, False otherwise."""
    if os.path.isdir(path):
        files = os.listdir(path)
        return len(files) == 0
    return not os.path.exists(path)


def check_same_num(dir_list: List[str]) -> bool:
    """Tells if all the folders in the list contain the same number of files."""
    num_files = [len([os.listdir(dir) for dir in dir_list])]
    if len(num_files) == 0:
        return True
    num_file = num_files[0]
    for element in num_files:
        if element != num_file:
            return False
    return True


def check_training(args: Namespace, logger: Logger) -> None:
    logger.info('Checking training pipeline was run correctly.')
    check_list = [
        os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'confs'),
        os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'normalizers'),
        os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'weights'),
        os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds'),
        os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds'),
        os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
    ]
    for element in check_list:
        if check_void(element):
            raise WrongConfigurationError(f'The following folder / file was not found: {element}. Try running the training pipeline again.')
    graph_check_list = [
        os.path.join(args.root_dir, 'gnn_logs', 'gnn_results.csv'),
        os.path.join(args.root_dir, 'gnn_logs', 'gcn_results.csv'),
        os.path.join(args.root_dir, 'gnn_logs', 'gat_results.csv'),
    ]
    exist_result_file = False
    graph_result_file = None
    for element in graph_check_list:
        if not check_void(element):
            exist_result_file = True
            graph_result_file = element
    if not exist_result_file:
        graph_folder = os.path.join(args.root_dir, 'gnn_logs')
        raise WrongConfigurationError(f'There is not graph result csv file under {graph_folder}')
        
    dir_list = [os.path.join(args.root_dir, 'weights', 'classification', 'gnn', dir) for dir in ['confs', 'normalizers', 'weights']]
    if not check_same_num(dir_list):
        raise WrongConfigurationError('Graph weights folders contain different number of files. Try running the training pipeline again')
    
    results = pd.read_csv(graph_result_file)
    if 'F1 Score' in results.columns and args.num_classes != 2:
        raise WrongConfigurationError('You have trained the GNN models for binary problem but are asking to run research for multiclass.')
    elif 'Micro F1' in results.columns and args.num_classes == 2:
        raise WrongConfigurationError('You have trained the GNN models for multiclass problem but are asking to run research for binary.')
        


import os
from typing import List, Dict, Any
from tqdm import tqdm
import argparse
from argparse import Namespace
from ..utils.preprocessing import get_names, read_gson, save_geojson


def process_gson(gson: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    return list(filter(lambda x: x['properties']['classification']['name'] != 'uncertain', gson))


def main_with_args(args: Namespace) -> None:
    os.makedirs(args.output_dir, exist_ok=True)
    names = get_names(args.input_dir, '.geojson')
    for name in tqdm(names):
        gson = read_gson(name, args.input_dir)
        out_gson = process_gson(gson)
        save_geojson(out_gson, name, args.output_dir)
    return


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input-dir', type=str, help='Folder where geojsons are with label uncertain.')
    parser.add_argument('--output-dir', type=str, help='Folder where to save geojsons without label uncertain.')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)


"""
Given .nodes.csv prediction files and .centroids.csv GT files,
replaces target value of .nodes.csv with GT ones.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pandas as pd
import numpy as np
from tqdm import tqdm
import argparse
from argparse import Namespace
import os
from ..utils.preprocessing import parse_path, create_dir, get_names, read_centroids, save_graph
from ..utils.nearest import generate_tree, find_nearest


def merge_labels(graph: pd.DataFrame, centroids: np.ndarray) -> pd.DataFrame:
    """
    Finds 1-1 matchings of nodes and substitutes labels in graph
    by those in centroids. Internally uses KD-trees.
    """
    assert len(centroids) > 0, 'GT must contain at least one cell.'
    graph = graph.copy()
    graph['prob1'] = graph['class'] - 1
    gt_tree = generate_tree(centroids[:,:2])
    pred_centroids = graph[['X', 'Y']].to_numpy(dtype=int)
    pred_tree = generate_tree(pred_centroids)
    for point_id, point in enumerate(centroids):
        if point[2] == -1:
            continue
        closest_id = find_nearest(point[:2], pred_tree)
        closest = graph.loc[closest_id, ['X', 'Y', 'class']]
        if closest[2] == -1:
            continue
        if point_id == find_nearest(closest[:2], gt_tree):
            graph.loc[closest_id, 'class'] = point[2] # 1-1 matchings
    for point_id, point in enumerate(pred_centroids):
        closest_id = find_nearest(point[:2], gt_tree)
        closest = centroids[closest_id]
        if point_id != find_nearest(closest[:2], pred_tree):
            graph.drop(point_id, inplace=True) # Remove prediction not matched in GT
    return graph


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--graph-dir', type=str, required=True,
        help='Path to folder containing prediction .nodes.csv files.'
    )
    parser.add_argument(
        '--centroids-dir', type=str, required=True,
        help='Path to folder containing .centroids.csv files.'
    )
    parser.add_argument(
        '--output-dir', type=str, required=True,
        help='Path to folder where to save new .nodes.csv. If same as --graph-dir, overwrites files.'
    )
    return parser


def main_with_args(args: Namespace) -> None:
    graph_dir = parse_path(args.graph_dir)
    centroids_dir = parse_path(args.centroids_dir)
    output_dir = parse_path(args.output_dir)
    create_dir(output_dir)
    names = sorted(get_names(centroids_dir, '.centroids.csv'))
    for name in tqdm(names):
        try:
            centroids = read_centroids(name, centroids_dir)
            graph = pd.read_csv(os.path.join(graph_dir, name + '.nodes.csv'))
            graph = merge_labels(graph, centroids)
            save_graph(graph, os.path.join(output_dir, name + '.nodes.csv'))
        except FileNotFoundError:
            continue
    return


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)


import argparse
from argparse import Namespace
from .preprocessing import geojson2pngcsv, pngcsv2graph, hovernet2geojson, pngcsv2centroids
from .segmentation import pngcsv2npy
import os
import logging
from logging import Logger
from .segmentation import hov_train, hov_infer
from .utils.preprocessing import get_names
import shutil
from .postprocessing import join_graph_gt, join_hovprob_graph
from .classification import train_gnn


def run_preproc_pipe(args: Namespace, logger : Logger) -> None:
    """
    Converts the gson format to the rest of formats.
    """
    for split in ['train', 'validation', 'test']:
        logger.info(f'Parsing {split} split')
        newargs = Namespace(
            gson_dir = os.path.join(args.root_dir, 'data', split, 'gson'),
            png_dir = os.path.join(args.root_dir, 'data', split, 'png'),
            csv_dir = os.path.join(args.root_dir, 'data', split, 'csv'),
            num_classes = args.num_classes,
        )
        geojson2pngcsv(newargs)
        newargs = Namespace(
            orig_dir = os.path.join(args.root_dir, 'data', 'orig'),
            png_dir = os.path.join(args.root_dir, 'data', split, 'png'),
            csv_dir = os.path.join(args.root_dir, 'data', split, 'csv'),
            out_dir = os.path.join(args.root_dir, 'data', split, 'npy'),
            save_example = False, use_labels = True, split = False,
            shape = '518'
        )
        pngcsv2npy(newargs)
    return


def run_hov_pipe(args: Namespace, logger : Logger) -> None:
    """
    Trains hovernet and predicts cell contours on json format.
    """
    logger.info('Starting training.')
    newargs = Namespace(
        gpu = args.gpu, view = None, save_name = None,
        log_dir = os.path.join(args.root_dir, 'weights', 'segmentation', 'hovernet'),
        train_dir = os.path.join(args.root_dir, 'data', 'train', 'npy'),
        valid_dir = os.path.join(args.root_dir, 'data', 'validation', 'npy'),
        pretrained_path = args.pretrained_path,
        shape = '518',
        num_classes = args.num_classes,
    )
    hov_train(newargs)
    logger.info('Starting inference.')
    newargs = {
        'nr_types': str(args.num_classes + 1),
        'type_info_path': os.path.join(args.root_dir, 'weights', 'segmentation', 'hovernet', 'type_info.json'),
        'gpu': args.gpu,
        'nr_inference_workers': '0',
        'model_path': os.path.join(args.root_dir, 'weights', 'segmentation', 'hovernet', '01', 'net_epoch=50.tar'),
        'batch_size': '10',
        'shape': '518',
        'nr_post_proc_workers': '0',
        'model_mode': 'original',
        'help': False
    }
    newsubargs = {
        'input_dir': os.path.join(args.root_dir, 'data', 'orig'),
        'output_dir': os.path.join(args.root_dir, 'data', 'tmp_hov'),
        'draw_dot': False,
        'save_qupath': False,
        'save_raw_map': False,
        'mem_usage': '0.2'
    }
    hov_infer(newargs, newsubargs, 'tile')
    return


def run_postproc_pipe(args: Namespace, logger : Logger) -> None:
    """
    Converts the json format to the graph format containing GT and preds.
    """
    logger.info('Moving json files to corresponding folders.')
    tr_files = set(get_names(os.path.join(args.root_dir, 'data', 'train', 'gson'), '.geojson'))
    val_files = set(get_names(os.path.join(args.root_dir, 'data', 'validation', 'gson'), '.geojson'))
    ts_files = set(get_names(os.path.join(args.root_dir, 'data', 'test', 'gson'), '.geojson'))
    json_files = set(get_names(os.path.join(args.root_dir, 'data', 'tmp_hov', 'json'), '.json'))
    for folder_name, split_files in zip(['train', 'validation', 'test'], [tr_files, val_files, ts_files]):
        for file in json_files.intersection(split_files):
            shutil.copy(
                os.path.join(args.root_dir, 'data', 'tmp_hov', 'json', file + '.json'),
                os.path.join(args.root_dir, 'data', folder_name, 'json')
            )
    for split in ['train', 'validation', 'test']:
        logger.info(f'Parsing {split} split')
        logger.info('   From json to geojson.')
        newargs = Namespace(
            json_dir = os.path.join(args.root_dir, 'data', split, 'json'),
            gson_dir = os.path.join(args.root_dir, 'data', split, 'gson_hov'),
            num_classes = args.num_classes
        )
        hovernet2geojson(newargs)
        logger.info('   From geojson to pngcsv.')
        newargs = Namespace(
            gson_dir = os.path.join(args.root_dir, 'data', split, 'gson_hov'),
            png_dir = os.path.join(args.root_dir, 'data', split, 'png_hov'),
            csv_dir = os.path.join(args.root_dir, 'data', split, 'csv_hov'),
            num_classes = args.num_classes,
        )
        geojson2pngcsv(newargs)
        logger.info('   From pngcsv to nodes.csv.')
        newargs = Namespace(
            png_dir = os.path.join(args.root_dir, 'data', split, 'png_hov'),
            csv_dir = os.path.join(args.root_dir, 'data', split, 'csv_hov'),
            orig_dir = os.path.join(args.root_dir, 'data', 'orig'),
            output_path = os.path.join(args.root_dir, 'data', split, 'graphs', 'raw'),
            num_workers = args.num_workers
        )
        pngcsv2graph(newargs)
        logger.info('   Extracting centroids from GT.')
        newargs = Namespace(
            png_dir = os.path.join(args.root_dir, 'data', split, 'png'),
            csv_dir = os.path.join(args.root_dir, 'data', split, 'csv'),
            output_path = os.path.join(args.root_dir, 'data', split, 'centroids')
        )
        pngcsv2centroids(newargs)
        logger.info('   Adding GT labels to .nodes.csv.')
        newargs = Namespace(
            graph_dir = os.path.join(args.root_dir, 'data', split, 'graphs', 'raw'),
            centroids_dir = os.path.join(args.root_dir, 'data', split, 'centroids'),
            output_dir = os.path.join(args.root_dir, 'data', split, 'graphs', 'GT')
        )
        join_graph_gt(newargs)
        logger.info('   Adding hovernet predictions to .nodes.csv.')
        newargs = Namespace(
            json_dir = os.path.join(args.root_dir, 'data', split, 'json'),
            graph_dir = os.path.join(args.root_dir, 'data', split, 'graphs', 'GT'),
            output_dir = os.path.join(args.root_dir, 'data', split, 'graphs', 'preds'),
            num_classes = args.num_classes,
        )
        join_hovprob_graph(newargs, logger)
    return


def run_graph_pipe(args: Namespace, logger : Logger) -> None:
    """
    Trains the graph models.
    """
    newargs = Namespace(
        train_node_dir = os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds'),
        validation_node_dir = os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds'),
        test_node_dir = os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
        log_dir = os.path.join(args.root_dir, 'gnn_logs'),
        early_stopping_rounds = 10,
        batch_size = 20,
        model_name = 'GCN',
        save_file = os.path.join(args.root_dir, 'gnn_logs', 'gcn_results'),
        num_confs = 32,
        save_dir = os.path.join(args.root_dir, 'weights', 'classification', 'gnn'),
        device = 'cpu' if args.gpu == '' else 'cuda',
        num_workers = args.num_workers,
        checkpoint_iters = -1,
        num_classes = args.num_classes,
        disable_prior = False,
        disable_morph_feats = False,
    )
    train_gnn(newargs)
    newargs = Namespace(
        train_node_dir = os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds'),
        validation_node_dir = os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds'),
        test_node_dir = os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
        log_dir = os.path.join(args.root_dir, 'gnn_logs'),
        early_stopping_rounds = 10,
        batch_size = 20,
        model_name = 'ATT',
        save_file = os.path.join(args.root_dir, 'gnn_logs', 'gat_results'),
        num_confs = 32,
        save_dir = os.path.join(args.root_dir, 'weights', 'classification', 'gnn'),
        device = 'cpu' if args.gpu == '' else 'cuda',
        num_workers = args.num_workers,
        checkpoint_iters = -1,
        num_classes = args.num_classes,
        disable_prior = False,
        disable_morph_feats = False,
    )
    train_gnn(newargs)
    return


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--root-dir', type=str, default='./.internals/', help='Root folder to save data and models.')
    parser.add_argument('--pretrained-path', type=str, help='Path to initial Hovernet weights.')
    parser.add_argument('--gpu', type=str, default='0')
    parser.add_argument('--num-workers', type=int, default=0)
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    return parser


def main():
    # TODO: Run each subpipe independently and measure time and memory requirements.

    parser = _create_parser()
    args = parser.parse_args()

    logger = logging.getLogger('train_pipe')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    logger.info('Starting preprocessing pipeline.')
    run_preproc_pipe(args, logger)
    logger.info('Finished preprocessing pipeline.')
    logger.info('Starting Hovernet pipeline.')
    run_hov_pipe(args, logger)
    logger.info('Finished Hovernet pipeline.')
    logger.info('Starting postprocessing pipeline.')
    run_postproc_pipe(args, logger)
    logger.info('Finished postprocessing pipeline.')
    logger.info('Starting graph pipeline.')
    run_graph_pipe(args, logger)
    logger.info('Finished graph pipeline.')



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import Callable, Tuple, List

Point = Tuple[float,float]
Contour = List[Point]
Cell = Tuple[int, int, Contour] # id, class

def rswoosh(
    I: List[Cell], 
    isEqual: Callable[[Contour,Contour], bool], 
    merge: Callable[[Cell, Cell], Cell]
    ) -> List[Cell]:
    """
    Given a list of cells, returns the same list without duplicates.
    Must provide a function to check equality and another to merge cells.
    """
    O = {} # Hash Table
    while(len(I) > 0):
        key_r, class_r, r = I.pop() # O(1)
        exist_equal = False
        for key_s, (class_s, s) in O.items():
            if isEqual(r,s):
                del O[key_s] # O(1)
                I.append(merge((key_r, class_r, r),(key_s, class_s, s))) # O(1)
                exist_equal = True
                break
        if not exist_equal:
            O[key_r] = (class_r, r) # O(1)
    return O.items()


import argparse
from argparse import Namespace
import logging
from logging import Logger
import numpy as np
import pandas as pd
from . import eval_segment
import os
from .preprocessing import hovernet2centroids, geojson2pngcsv, pngcsv2centroids
from .utils.preprocessing import get_names
from .utils.pipes import HovernetNotFoundError, check_void
from .utils.classification import metrics_from_predictions


def run_preprocessing(args: Namespace, logger : Logger) -> None:
    """
    Converts the gson format to the rest of formats.
    """
    logger.info('Extracting Hovernet centroids from training output.')
    if not os.path.isdir(os.path.join(args.root_dir, 'data', 'tmp_hov', 'json')):
        raise HovernetNotFoundError('Please, train again or extract hovernet outputs.')
    newargs = Namespace(
        json_dir = os.path.join(args.root_dir, 'data', 'tmp_hov', 'json'),
        output_path = os.path.join(args.root_dir, 'data', 'tmp_hov', 'centroids_hov'),
    )
    hovernet2centroids(newargs)
    for split in ['train', 'validation', 'test']:
        if check_void(os.path.join(args.root_dir, 'data', split, 'names.txt')):
            logger.info(f'Preprocessing split {split}.')
            split_names = get_names(os.path.join(args.root_dir, 'data', split, 'gson'), '.geojson')
            with open(os.path.join(args.root_dir, 'data', split, 'names.txt'), 'w') as f:
                for name in split_names:
                    print(name, file=f)
        if check_void(os.path.join(args.root_dir, 'data', split, 'png')) or check_void(os.path.join(args.root_dir, 'data', split, 'csv')):
            logger.info('   From geojson to pngcsv.')
            newargs = Namespace(
                gson_dir = os.path.join(args.root_dir, 'data', split, 'gson'),
                png_dir = os.path.join(args.root_dir, 'data', split, 'png'),
                csv_dir = os.path.join(args.root_dir, 'data', split, 'csv'),
                num_classes = args.num_classes,
            )
            geojson2pngcsv(newargs)
        if check_void(os.path.join(args.root_dir, 'data', split, 'centroids')):
            logger.info('   Extracting centroids from GT.')
            newargs = Namespace(
                png_dir = os.path.join(args.root_dir, 'data', split, 'png'),
                csv_dir = os.path.join(args.root_dir, 'data', split, 'csv'),
                output_path = os.path.join(args.root_dir, 'data', split, 'centroids')
            )
            pngcsv2centroids(newargs)
    return


def compute_ece(args: Namespace, logger: Logger, split: str) -> None:
    folder = os.path.join(args.root_dir, 'data', split, 'graphs', 'preds')
    trues, probs = None, None
    for file in os.listdir(folder):
        df = pd.read_csv(os.path.join(folder, file))
        _trues = df['class'].to_numpy()-1
        if trues is None:
            trues = _trues.reshape((-1,1))
        else:
            trues = np.vstack((trues, _trues.reshape((-1, 1))))
        if args.num_classes == 2:
            _probs = df['prob1'].to_numpy().reshape((-1, 1))
            _probs = np.hstack((1 - _probs, _probs))
        else:
            cols = ['prob' + str(k) for k in range(1, args.num_classes + 1)]
            _probs = df[cols].to_numpy()
        if probs is None:
            probs = _probs
        else:
            probs = np.vstack((probs, _probs))
    preds = np.argmax(probs, axis=1).reshape((-1, 1))
    metrics = metrics_from_predictions(trues, preds, probs[:, 1], args.num_classes)
    if args.num_classes == 2:
        acc, f1, auc, perc_error, ece = metrics
        dic_metrics = {
            'F1': [f1], 'Accuracy': [acc], 'ROC_AUC': [auc], 'Perc_err': [perc_error], 'ECE': [ece]
        }
    else:
        micro, macro, weighted, ece = metrics
        dic_metrics = {
            'Macro F1': [macro], 'Weighted F1': [weighted], 'Micro F1': [micro], 'ECE': [ece]
        }
    metrics_df = pd.DataFrame(dic_metrics)
    metrics_df.to_csv(args.save_name + '_' + split + '_ece.csv', index=False)


def run_evaluation(args: Namespace, logger: Logger) -> None:
    logger.info('Starting evaluation of Hovernet output.')
    for split in ['train', 'validation', 'test']:
        logger.info(f'    Evaluating {split} split')
        newargs = Namespace(
            names = os.path.join(args.root_dir, 'data', split, 'names.txt'),
            gt_path = os.path.join(args.root_dir, 'data', split, 'centroids'),
            pred_path = os.path.join(args.root_dir, 'data', 'tmp_hov', 'centroids_hov'),
            save_name = args.save_name + '_' + split,
            debug_path = None,
            num_classes = args.num_classes
        )
        eval_segment(newargs, logger)

        compute_ece(args, logger, split)
    return


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--root-dir', type=str, default='./.internals/', help='Root folder to save data and models.')
    parser.add_argument('--save-name', type=str, required=True, help='Name to save the result, without file type.')
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()

    logger = logging.getLogger('eval_pipe')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    run_preprocessing(args, logger)
    run_evaluation(args, logger)